/*
** pork.c - main function.
** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <pwd.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <libfaim/aim.h>
#include <libfaim/faimconfig.h>

#include <pork.h>
#include <pork_util.h>
#include <pork_list.h>
#include <pork_buddy.h>
#include <pork_set.h>
#include <pork_swindow.h>
#include <pork_imwindow.h>
#include <pork_slist.h>
#include <pork_buddy_list.h>
#include <pork_acct.h>
#include <pork_aim.h>
#include <pork_misc.h>
#include <pork_input.h>
#include <pork_bind.h>
#include <pork_events.h>
#include <pork_screen.h>
#include <pork_screen_cmd.h>
#include <pork_status.h>
#include <pork_alias.h>
#include <pork_opt.h>
#include <pork_conf.h>
#include <pork_color.h>
#include <pork_command.h>
#include <pork_perl.h>
#include <pork_timer.h>
#include <pork_io.h>

struct screen screen;

/*
** The fallback for when no binding for a key exists:
**
** Write the key to the input buffer.  If the key is a control
** character, output ^<KEY> to avoid messing up the display.
*/

static void binding_insert(int key) {
	struct imwindow *imwindow = cur_window();

	if (key <= 0xff) {
		if (iscntrl(key)) {
			static char buf[3] = { '^', '\0', '\0' };

			buf[1] = toupper(key) + 'A' - 1;
			input_insert_str(imwindow->input, buf);
		} else
			input_insert(imwindow->input, key);
	}
}

static inline void binding_run(struct binding *binding) {
	/*
	** Yeah, this is kind of a hack, but it makes things
	** more pleasant (i.e. you don't see status messages
	** about new settings and things like that if the command
	** was run from a key binding (other than hitting enter on
	** the command line).
	*/

	if (binding->key != '\n') {
		int quiet = screen_set_quiet(1);

		run_mcommand(binding->binding);
		screen_set_quiet(quiet);
	} else
		run_mcommand(binding->binding);
}

static void sigwinch_handler(int sig __notused) {
	struct winsize size;

	if (ioctl(1, TIOCGWINSZ, &size) != 0)
		pork_exit(-1, "Fatal error getting screen size\n");

	if (size.ws_row < 3 || size.ws_col < 2)
		pork_exit(-1, "You made the screen too small.  Don't do that\n");

	resizeterm(size.ws_row, size.ws_col);
	screen_resize(size.ws_row, size.ws_col);
	screen_refresh();
}

#if 0
static void sigsegv_handler(int sig __notused) {
	endwin();
	fprintf(stderr, "You found a bug.\n");
	exit(-1);
}
#endif

static void generic_signal_handler(int sig) {
	pork_exit(sig, "Caught signal %d.  Exiting\n", sig);
}

void keyboard_input(int fd __notused,
					u_int32_t cond __notused,
					void *data __notused)
{
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	int key;

	key = wgetinput(screen.status_bar);
	if (key == -1)
		return;

	time(&priv->last_input);

	if (priv->connected && priv->marked_idle && opt_get_bool(OPT_REPORT_IDLE)) {
		screen_status_msg("%s is no longer marked idle", priv->username);
		aim_set_idle(priv, 0);
		priv->marked_idle = 0;
	}

	bind_exec(imwindow->active_binds, key);
}

int main(int argc, char **argv) {
	struct passwd *pw;
	char buf[PATH_MAX];
	struct imwindow *imwindow;
	int ret;
	time_t timer_last_run;
	time_t status_last_update = 0;

    pw = getpwuid(getuid());
    if (pw == NULL) {
		fprintf(stderr, "Fatal: Can't get your user info.\n");
        exit(-1);
	}

	snprintf(buf, sizeof(buf), "%s/.pork", pw->pw_dir);
	opt_set(OPT_PORK_DIR, buf);

	perl_init();

	if (get_options(argc, argv) != 0) {
		fprintf(stderr, "Fatal: Error getting options.\n");
		exit(-1);
	}

	if (initialize_environment() != 0) {
		fprintf(stderr, "Fatal: Error initializing the terminal.\n");
		exit(-1);
	}

	init_colors();
	pork_io_init();

	if (screen_init(LINES, COLS) == -1)
		pork_exit(-1, "Fatal: Error initializing the terminal.\n");

	signal(SIGWINCH, sigwinch_handler);
	signal(SIGTERM, generic_signal_handler);
	signal(SIGQUIT, generic_signal_handler);
	signal(SIGHUP, SIG_IGN);
//	signal(SIGSEGV, sigsegv_handler);

	wmove(screen.status_bar, STATUS_ROWS - 1, 0);
	imwindow = cur_window();

	bind_init(&screen.binds);
	bind_set_handlers(&screen.binds.main, binding_run, binding_insert);
	bind_set_handlers(&screen.binds.blist, binding_run, NULL);
	alias_init(&screen.alias_hash);
	events_init(&screen.events);

	screen_set_quiet(1);
	ret = read_global_config();
	screen_set_quiet(0);

	if (ret != 0)
		screen_err_msg("Error reading the global configuration.");

	status_draw(&screen.null_priv);
	screen_draw_input();
	screen_doupdate();

	time(&timer_last_run);
	while (1) {
		time_t time_now;

		pork_io_run();
		pork_acct_update();

		/*
		** Run the timers at most once per second.
		*/

		time(&time_now);
		if (timer_last_run < time_now) {
			timer_last_run = time_now;
			timer_run(&screen.timer_list);
		}

		imwindow = cur_window();
		imwindow_refresh(imwindow);

		/*
		** Draw the status bar at most once per second.
		*/

		time(&time_now);
		if (status_last_update < time_now) {
			status_last_update = time_now;
			status_draw(imwindow->priv);
		}

		screen_draw_input();
		screen_doupdate();
	}

	pork_exit(0, NULL);
}

/*
** Sign off all acounts and exit.  If a message
** is given, print it to the screen.
*/

void pork_exit(int status, char *fmt, ...) {
	pork_acct_del_all();
	screen_destroy();
	perl_destroy();
	pork_io_destroy();

	wclear(stdscr);
	wrefresh(stdscr);
	delwin(stdscr);
	endwin();

	if (fmt != NULL) {
		va_list ap;

		va_start(ap, fmt);
		vfprintf(stderr, fmt, ap);
		va_end(ap);
	}

	exit(status);
}
